//#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <malloc.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>
#include <conio.h>

#define TIME_IN_MS
#define MAX_MSG_SIZE    128
#define MAX_NUM_COPIES  700                 /* Upper Limit of number of "children" */

/*      MONKEY  Version 1.0

    This program uses cumulative selection to produce a target string
(e.g. the alphabet in alphabetical order) through random changes.  For more
information on this subject, read the text files, MONKEY.DOC and MPROBS.DOC,
the source code of MONKEY, and Chapter 3 of Richard Dawkins' book, _The Blind
Watchmaker_.
*/


// CONST
  const double year = 365.2425 * 24.0 * 3600.0;   /* year in seconds */
  const char *test_msg = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";  /* Alphabet test message */
  const int alphabet_pool_size = 26;          /* number of letters used in alphabet test */
  const int user_pool_size = 27;               /* number of letters in user-entered test */
  /* valid characters for user-entered test (invalid characters get converted to ' ') */
  const char valid_set[] = {' ','A','B','C','D','E','F','G','H','I','J','K',
                            'L','M','N','O','P','R','S','T','U','V','W','X',
                            'Y','Z','a','b','c','d','e','f','g','h','i','j',
                            'k','l','m','n','o','p','q','r','s','t','u','v'
                            ,'w','x','y','z'};

/* Pool of letters picked from randomly (' ' not used in alphabet test) */
  const char letter_pool[] = {'A','B','C','D',
                      'E','F','G','H','I','J','K','L','M','N','O','P',
                      'Q','R','S','T','U','V','W','X','Y','Z',' '};

// TYPE
  typedef long double real_type; /* = Extended; */              /* 8087 type:  3.4E-4932 .. 1.1E4932 */
typedef enum {alphabet,entered} target_type;   /* Indicates whether target string is the
                                        alphabet or a user-entered string */
//  typedef char string80[81];              /* the string type */
  
#ifndef TIME_IN_MS
  typedef struct                    /* Time Record Type */
  {
    int hour;
    int min;
    int sec;
    int sec100;
  } TIME_REC;    
#endif TIME_IN_MS

// VAR
#ifdef TIME_IN_MS
  clock_t start_time;                         /* Start and Stop times and duration  */
  clock_t stop_time;                          /*   for measuring elapsed time       */
  clock_t duration;
#else  TIME_IN_MS
  TIME_REC start_time;                         /* Start and Stop times and duration  */
  TIME_REC stop_time;                          /*   for measuring elapsed time       */
  TIME_REC duration;
#endif TIME_IN_MS
  double tries;                           /* Reciprocal of single-step probability */
  double years; //      :real_type;        /* time in years to make all possible tries */
  target_type target;      /* indicates type of target string */
  int option;                       /* single-step or cumulative selection (all
                                    change or all save one) */
  int method;             /* change single letter or test each letter
                                    position to change */
  int pool_size;             /* current size of letter pool */
  int num_copies;          /* selected number of offspring per generation */
  double mut_prob;             /* probability that a given letter position
                                    will change */
  int num_changes;            /* number of changes in each offspring */
  int quit_flag;          /* indicates selection of QUIT PROGRAM option */

  char msg[MAX_MSG_SIZE+1];                          /* Target String */
  char user_msg[MAX_MSG_SIZE+1];         /* User entered string */
  int  msg_size;             /* Length of Target String */

char parent[MAX_MSG_SIZE+1] = "";
char offspring[MAX_NUM_COPIES][MAX_MSG_SIZE+1] = {""};

//  string80 s[MAX_NUM_COPIES+1];  /* Mutated copies */
int  diffs[MAX_NUM_COPIES];   

/* Differences between the
                                                       copies and the target */
// { Constants and Variables for LOG and POWER }

// CONST
#define MAX_VAL 1.0E30       //  { Within range of REAL type }
#define MIN_VAL 1.0E-30
#define BIG_E MAX_VAL
#define LITTLE_E MIN_VAL

double big_ln,               //    { LN(max_val)  }
       little_ln;        // { LN(min_val)  }


void spawn(void);
void single_selection(void);
void cumulative_selection(void);

void get_options(void);
#ifdef TIME_IN_MS
void get_elapsed_time(clock_t start, clock_t *duration);
void report(long loops, clock_t duration);
#else  TIME_IN_MS
void get_elapsed_time(TIME_REC *start, TIME_REC *duration);
void report(long loops, TIME_REC *duration);
#endif TIME_IN_MS

void ReportSingleStep(int iy, unsigned long loops, int closest_diff, 
                        unsigned long closest_loop);

void get_target(void);
void setup_screen(void);
int calc_diff(char *s);
int min_diff(int *diffs);
void initialize_string(char *parent, int *diff);

int Random(int n);
double random(void);
real_type power(real_type v, int expon);

int main(int argc, char *argv[])
{

// {  Initialize values for LOG and POWER  }

    big_ln = log(BIG_E);
    little_ln = log(LITTLE_E);

// {  Initialize options to default to the Alphabet Test  }

    option = 2;
    target = alphabet;
    strcpy(user_msg,"");
    method = 1;
    num_copies = 100;
    mut_prob = 0.1;
    num_changes = 2;

//  { Main Loop }

    srand((unsigned)time(NULL));
    do
    {
        get_options();
        if (!quit_flag)
        {
            if (option == 1)
                single_selection();
            else
                cumulative_selection();
        }
    } while (!quit_flag);
  
    return 0;
}

// (* THE MAIN ROUTINES *)

#ifndef TIME_IN_MS
    struct tm *start;
    time_t  tijd;
#endif TIME_IN_MS

void StartTheClock()
{
#ifdef TIME_IN_MS
    start_time = clock();
#else  TIME_IN_MS
    tijd = time(NULL);
    start = localtime(&tijd);
    start_time.hour = start->tm_hour;
    start_time.min = start->tm_min;
    start_time.sec = start->tm_sec;
    start_time.sec100 = 0;
#endif TIME_IN_MS
}

void StopTheClock()
{
    get_elapsed_time(start_time,&duration);
}


void SetupSingleSelectionScreen(int *iy)
{
    int y;
    
    
    setup_screen();
    y = wherey();
    gotoxy(1, y);
    printf("Single-Step Selection:");
    gotoxy(3, y+2);
    printf("Attempts:    ");
    gotoxy(3, y+6);
    printf("Closest:  differ by     after             attempts.");
    *iy = y;
}

void UpdateSingleSelectionScreen(int iy, unsigned long loops, char *str, 
                    int diff, int closest_diff, unsigned long closest_loop)
{
        gotoxy(13, iy+2);
        printf("%lu.",loops);
        gotoxy(3, iy+4);
        clreol();
        printf("%s %6d",str,diff);
        gotoxy(23, iy+6);
        printf("%d",closest_diff);
        gotoxy(34, iy+6);
        printf("%lu", closest_loop);
}

/*  SINGLE_SELECTION tries to generate the target string through single-step
   selection (i.e. repeatedly try to generate the target string from scratch).
   After setting up the screen and intializing the variables, control enters
   into a REPEAT loop which increments the loop counter, generates the random
   string, and compares it to the target string.  The smallest difference,
   i.e. the closest approach to the target, is retained in CLOSEST for display
   upon exiting the loop.  Control exits the loop either when the target
   string has been generated or the user presses the ALT key.  The results are
   reported and control returns to the Main Menu.   */
void single_selection(void)
{
    int  diff, closest_diff;
    unsigned long loops, closest_loop;
    int  exit = 0;
    char *str;
    int  iy;
    char ch;
    
    str = malloc(msg_size * sizeof(char)+1);
    
    get_target();
    closest_diff = INT_MAX;
    loops = 0;
    closest_loop = loops;
    SetupSingleSelectionScreen(&iy);
    StartTheClock();
    do 
    {
        loops++;
        initialize_string(str, &diff);
        if (closest_diff > diff)
        {
            closest_diff = diff;
            closest_loop = loops;
        }
        UpdateSingleSelectionScreen(iy, loops, str, diff, closest_diff, closest_loop);
        if (kbhit())
            exit = 1;            
    } while (diff && !exit); // UNTIL (diffs[0] = 0) OR ((MEM[0000:$0417] AND 8) <> 0);

    StopTheClock();
    ReportSingleStep(iy, loops, closest_diff, closest_loop);

    gotoxy(1,25);
    clreol();
    printf("Press space bar to continue ");
    do
    {
        ch = getch();
    } while (ch != ' ');
}



/*  SPAWN copies the "parent" string (index 0) into each "offspring" string and
   then performs the selected "mutation".

    NOTE:  in each case, the letter position to be changed is selected at
           random AND the letter to be placed there is also selected at
           random.   */
void spawn()
{
    int i, j, k;

    for (i=0; i < num_copies; i++)
    {
        strcpy(offspring[i], parent);                 /* Copy the "parent" */
        if ( (option != 3) )   /* Do not change if first child and option */
        {
            switch(method)   /*    is 3 (one child remains unchanged) */
            {
                case 1 : 
                    offspring[i][(Random(msg_size))] = letter_pool[Random(pool_size)];
                    break;
                case 2 : 
                    for (k=0; k<num_changes; k++)
                        offspring[i][(Random(msg_size))] = letter_pool[Random(pool_size)];
                    break;
                case 3 : 
                    for (j=0; j<msg_size; j++)
                         // NEED TO DETERMINE Random CALL
                         // X := Random    { 0.0 <= X < 1.0 } ;
                         // J := Random(N) {  0  <= J <  N  } ;

                        if  (random() < mut_prob)
                            offspring[i][j] = letter_pool[Random(pool_size)];
            }  /*CASE*/
        }
        diffs[i] = calc_diff(offspring[i]);
    }  /*FOR i:=1 TO num_copies DO*/

}



void SetupCumulativeSelectionScreen(int *iy)
{
    int y;
    
    
    setup_screen();
    y = wherey();
    gotoxy(1, y);
    printf("Single-Step Selection:");
    gotoxy(3, y+2);
    printf("Attempts:    ");
    gotoxy(3, y+6);
    printf("Closest:  differ by     after             attempts.");
    *iy = y;
}

void UpdateCumulativeSelectionScreen(int iy, unsigned long loops, char *str, 
                    int diff, int closest_diff, unsigned long closest_loop)
{
        gotoxy(13, iy+2);
        printf("%lu.",loops);
        gotoxy(3, iy+4);
        clreol();
        printf("%s %6d",str,diff);
        gotoxy(23, iy+6);
        printf("%d",closest_diff);
        gotoxy(34, iy+6);
        printf("%lu", closest_loop);
}


/*  CUMULATIVE_SELECTION tries to generate the target string through cumulative
   selection (i.e. using a parent string, generate multiple copies which differ
   slightly from the parent, and use the copy that comes closest to the target
   as the next parent).
   After setting up the screen and intializing the variables, control enters
   into a REPEAT loop which increments the loop counter and spawns the next
   generation.  The new parent string and its difference from the target
   string are displayed each time.  Control exits the loop either when the
   target string has been generated or the user presses the ALT key at which
   time the results are reported.  The user then chooses either to run the
   same test again or to return to the Main Menu.   */

void cumulative_selection(void)
{
#if 0
    int     i;
    long  loops;
    char  ch;
    int   iy;
    int   exit = 0;
    int  diff;
    
    do 
    {
        get_target();
        setup_screen();
        initialize_string(parent, &diff);
        iy = wherey();
        loops = 0;

        StartTheClock();

        do
        {
            loops++;
            spawn();
            i = min_diff(diffs);                  /* find offspring closest to target */
            strcpy(parent, offspring[i]);                   /* make the offspring the next parent */
            diff = diffs[i];           /* save its difference for display  */

            gotoxy(1,iy);
            printf("Iterations:  %ld\n",loops);
            printf("\n");
            printf("\n");
            printf("Differs by: %4d\n",diff);
            if (kbhit())
                exit = 1;            
        } while (diff && !exit); // UNTIL (diffs[0] = 0) OR ((MEM[0000:$0417] AND 8) <> 0);

        StopTheClock();
        ReportSingleStep(iy, loops, closest_diff, closest_loop);

        gotoxy(1,25);
        clreol();
        printf("Press Space Bar to Exit; Press any other key to Repeat");
        ch = getch();
        printf("\n");
    } while (ch != ' ');
#else  // ----------------------------------
    int     i;
    long  loops;
    char  ch;
    int   iy;
#ifndef TIME_IN_MS
    struct tm *start;
    time_t  tijd;
#endif  TIME_IN_MS
    int   exit = 0;
    int  diff;
    
    do 
    {
        get_target();
        setup_screen();
        initialize_string(parent, &diff);
        iy = wherey();
        loops = 0;

#ifdef TIME_IN_MS
        start_time = clock();
#else  TIME_IN_MS
        tijd = time(NULL);
        start = localtime(&tijd);
        start_time.hour = start->tm_hour;
        start_time.min = start->tm_min;
        start_time.sec = start->tm_sec;
        start_time.sec100 = 0;
#endif TIME_IN_MS

        do
        {
            loops++;
            spawn();
            i = min_diff(diffs);                  /* find offspring closest to target */
            strcpy(parent, offspring[i]);                   /* make the offspring the next parent */
            diff = diffs[i];           /* save its difference for display  */

            gotoxy(1,iy);
            printf("Iterations:  %ld\n",loops);
            printf("\n");
     //       printf("%s\n",parent);
            printf("\n");
            printf("Differs by: %4d\n",diff);
            if (kbhit())
                exit = 1;            
        } while (diff && !exit); // UNTIL (diffs[0] = 0) OR ((MEM[0000:$0417] AND 8) <> 0);
#ifdef TIME_IN_MS
        get_elapsed_time(start_time,&duration);
        printf("\n");
        report(loops,duration);
#else  TIME_IN_MS
        get_elapsed_time(&start_time,&duration);
        printf("\n");
        report(loops,&duration);
#endif TIME_IN_MS
//        WINDOW(1,1,80,25);
        gotoxy(1,25);
        clreol();
        printf("Press Space Bar to Exit; Press any other key to Repeat");
        ch = getch();
        printf("\n");
    } while (ch != ' ');
#endif    
}


// (* MAN-MACHINE INTERFACE (MMI) *)




/*  GET_OPTIONS presents the Main Menu in which the user can select the
   parameters for the test.  Single key-stroke entry of the item number
   steps the selected item to its next value.  Once all parameters are
   set, pressing "X" starts the test.  Pressing "Q" terminates the program. */

void get_options(void)
{
    int   i, iy;
    char  ch, sel;
    char  sel_set[40] = "";
    int   flag, go_flag;

    go_flag = 0;
    quit_flag = 0;
    do
    {
        clrscr();
        printf("MONKEY\n");
        printf("\n");
        printf("1. Target:  ");
        if (target == alphabet)
            printf("Alphabet\n");
        else
        {
            printf("User Entered -- \n");
            printf("\"%s\"\n",user_msg);
        }
        printf("\n");
        printf("2. Option:  ");
        switch (option)
        {
            case 1 : 
                printf("Single-Step Selection\n");
                break;
            case 2 : 
                printf("Cumulative Selection -- All Offspring Change\n");
                break;
            case 3 : 
                printf("Cumulative Selection -- One Offspring Does Not Change\n");
                break;
        }
        printf("\n");
        strcpy(sel_set, "12XQ");
        if (option != 1)
        {
            printf("3. Method:  ");
            switch (method)
            {
                case 1 : 
                    printf("Change Single Letter\n");
                    break;
                case 2 : 
                    printf("Change Multiple Letters\n");
                    break;
                case 3 : 
                    printf("Chance of Changing Any or All Letters\n");
                    break;
            }
            printf("\n");
            printf("4. Number of Offspring:  %d\n",num_copies);
            printf("\n");
            strcat(sel_set, "34");
            switch (method)
            {
                case 2 :
                    printf("5. Number of Changes per Offspring:  %8d\n",num_changes);
                    printf("\n");
                    strcat(sel_set, "5");
                    break;
                case 3 :
                    printf("5. Base Probability of Change:  %8.4f\n",mut_prob);
                    printf("\n");
                    strcat(sel_set, "5");
                    break;
                default:
                   ; // { do nothing }
            }
        }
        printf("\n");
        printf("X.  Exit (and Run Program)\n");
        printf("\n");
        printf("Q.  Quit\n");
        printf("\n");
        printf("\n");
        printf("Enter Choice:  ");
        do 
        {
            sel = getch();
        } while (strchr(sel_set,sel) == NULL);
        gotoxy(1,wherey());
        clreol();
        switch (sel)
        {
            case '1' : 
                if (target == alphabet)
                {
                    target = entered;
                    iy = wherey();
                    gotoxy(1,iy);
                    if (strlen(user_msg) == 0)
                    {
                        printf("No Message has been Entered\n");
                        printf("\n");
                        flag = 1;
                    }
                    else
                    {
                        printf("Message is:\n");
                        printf("\"%s\"\n",user_msg);
                        printf("Do You Wish to Enter a Different Message? (Y/N) ");
                        do
                        {
                            ch = toupper(getche());
                        } while (ch != 'Y' && ch != 'N');
                        flag = (ch == 'Y');
                    }
                    if (flag)
                    {
                        gotoxy(1,wherey());
                        clreol();
                        printf("Enter New Message:\n");
                        scanf("%s\n", user_msg);
                        for (i=0; i<strlen(user_msg); i++)  // { Convert user string }
                            if (strchr(valid_set,user_msg[i]) != NULL) //{ to CAPs and blank   }
                                user_msg[i] = toupper(user_msg[i]);    //{ out      }
                            else                                   // { invalid characters  }
                                user_msg[i] = ' ';
                    }
                }
                else
                    target = alphabet;
            case '2' : 
                option++;
                if (option > 3)
                    option = 1;
                break;
            case '3' :
                method++;
                if (method > 3)
                    method = 1;
                break;
            case '4' :
                gotoxy(10,20);
                printf("Enter Number of Offspring:  ");
                scanf("%d\n", &num_copies);
                break;
            case '5' : 
                switch (method)
                {
                    case 2:
                        gotoxy(10,wherey());
                        printf("Enter Number of Changes per Offspring:  ");
                        scanf("%d\n", &num_changes);
                        break;
                    case 3:
                        gotoxy(10,wherey());
                        printf("Enter Base Probability for Change:  ");
                        scanf("%lf\n", &mut_prob);
                        break;
                    default:
                        ; // {do nothing}
                } //END;  {CASE}
                break;
            case 'Q' : 
                quit_flag = 1;
                break;
            case 'X' : 
                go_flag = 1;
                break;
        } // END;  {CASE}
    } while (!go_flag && !quit_flag);
}



/*  GET_ELAPSED_TIME reads the stop time and calculates the elapsed time in
   hours, minutes, seconds, and 1/100's of seconds.  */

void adjust_time(int *hi_unit, int *lo_unit, int factor)
{
    if (lo_unit < 0)
    {
        *lo_unit += factor;
        (*hi_unit)--;
    }
}

#ifdef TIME_IN_MS
void get_elapsed_time(clock_t start, clock_t *duration)
{
    *duration = clock() - start;
    if (*duration == 0)
        *duration = 1;
}
#else  TIME_IN_MS
void get_elapsed_time(TIME_REC *start, TIME_REC *duration)
{
    struct tm *stop_time;
    time_t  tijd;
    
    tijd = time(NULL);
    stop_time = localtime(&tijd);
    
    duration->hour = stop_time->tm_hour - start->hour;
    duration->min = stop_time->tm_min - start->min;
    duration->sec = stop_time->tm_sec - start->sec;
    duration->sec100 = 1; //stop_time->tm_sec100 - start->sec100;
    adjust_time(&duration->sec,&duration->sec100,100);
    adjust_time(&duration->min,&duration->sec,60);
    adjust_time(&duration->hour,&duration->min,60);
}
#endif TIME_IN_MS



/*  REPORT displays the statistics of the run:  number of iterations, the
   amount of time test had run, and the equivalent number of tries and time
   required in years if single-step selection had been used.             */

#ifdef TIME_IN_MS
void report(long loops, clock_t duration)
{
    int hour, min, sec, msec;
    double  dTime = (double)duration / CLOCKS_PER_SEC;
    
    hour = (int)(dTime/3600.0);
    dTime -= hour * 3600.0;
    min = (int)(dTime/60.0);
    dTime -= min * 60.0;
    sec = (int)dTime;
    dTime -= sec;
    msec = (int)(dTime*1000.0 + 0.5);
    
    printf("\nCumulative Selection Method succeeded with:\n");
    printf("%ld iterations in %2d:%02d:%02d.%03d\n", loops, hour, min, sec, msec);
    tries = power(pool_size,msg_size);
    dTime = (double)duration / CLOCKS_PER_SEC;
    years = dTime / year;
    years = years * tries / loops;
    printf("\n");
    printf("Single Step Method would have required:\n");
    printf("%e attempts in %e years\n", tries, years);
    printf("\nEstimated Age of the Universe: %e years\n", 13e+9);
}
#else  TIME_IN_MS
void report(long loops, TIME_REC *duration)
{
    printf("\nCumulative Selection Method succeeded with:\n");
    printf("%ld iterations in %2d:%02d:%02d.%02d\n", loops, duration->hour,
            duration->min,duration->sec,duration->sec100);
    tries = power(pool_size,msg_size);
    years = (duration->hour * 3600.0 + duration->min * 60.0 + duration->sec + 
                duration->sec100 / 100) / year;
    years = years * tries / loops;
    printf("\n");
    printf("Single Step Method would have required:\n");
    printf("%e attempts in %e years\n", tries, years);
    printf("\nEstimated Age of the Universe: %e years\n", 13e+9);
}
#endif TIME_IN_MS


void ReportSingleStep(int iy, unsigned long loops, int closest_diff, unsigned long closest_loop)
{
#ifdef TIME_IN_MS
    int hour, min, sec, msec;
    double  dTime = (double)duration / CLOCKS_PER_SEC;
    
    hour = (int)(dTime/3600.0);
    dTime -= hour * 3600.0;
    min = (int)(dTime/60.0);
    dTime -= min * 60.0;
    sec = (int)dTime;
    dTime -= sec;
    msec = (int)(dTime*1000.0 + 0.5);
#else
    double dTime;
#endif TIME_IN_MS

    gotoxy(1, iy+9);
    printf("Closest Approach = %d after %lu attempts.\n",closest_diff, closest_loop);
    gotoxy(1, iy+11);
#ifdef TIME_IN_MS
    printf("%lu iterations in %2d:%02d:%02d.%03d\n", loops, hour, min, sec, msec);
    dTime = (double)duration / CLOCKS_PER_SEC;
#else  TIME_IN_MS
    printf("%ld iterations in %2d:%02d:%02d.%02d\n", loops, duration->hour,
            duration->min,duration->sec,duration->sec100);
    dTime = duration->hour * 3600.0 + duration->min * 60.0 + duration->sec + 
                duration->sec100 / 100;
#endif TIME_IN_MS
    gotoxy(5, iy+12);
    printf("Average %g attempts per second.\n", (double)loops/dTime);

    years = dTime / year;
    tries = power(pool_size,msg_size);
    years = years * tries / loops;
    gotoxy(1, iy+15);
    printf("Single Step Method would have required:\n");
    printf("    %e attempts in %e years for one-in-a-million odds.\n", tries, years);
    printf("\n  Estimated Age of the Universe: %e years\n", 13e+9);
}


/*
Single-Step Selection:

  Attempts: 1028867.
 QUVOMCZEJQVIYHVGFRUYHYXIPV     25


  Closest:  differ by 18  after  15890      attempts.


Closest Approach = 18 after 15890 attempts.

1028867 iterations in  0:03:28.045

Single Step Method would have required:
    6.156120e+036 attempts in 3.944664e+025 years for one-in-a-million odds.

  Estimated Age of the Universe: 1.300000e+010 years







Press space bar to continue
*/

//(* PROGRAM SUPPORT *)



/*  GET_TARGET sets the target string to either the alphabet or to the
   user-entered string.  If the alphabet test is being run, then only the
   characters "A" to "Z" are used.  If the user-entered string is used, then
   blanks are also used. */

void get_target(void)
{
    if (target == alphabet)
    {
        strcpy(msg, test_msg);
        pool_size = alphabet_pool_size;
    }
    else
    {
        strcpy(msg, user_msg);
        pool_size = user_pool_size;
    }
  msg_size = strlen(msg);
}




/*  SETUP_SCREEN clears the screen and displays the target string with its
   length at the top and instructions to exit at the bottom.  */

void setup_screen(void)
{
  clrscr();
  printf("\"%s\"\n", msg);
  printf("%d\n", msg_size);
  gotoxy(1,25);
  printf("Press ALT key to Exit");
//  window(1,4,80,23);
  clrscr();
}



/*  DIFF measures the difference between the string S and the target by two
   means.  First, a count is kept of how many letters in S are different from
   the corresponding letter in the target.  If all letters are different, then
   a summation of the letter-by-letter difference between S and the target is
   calculated.

   NOTE:  Only in the single-step test have I ever seen the summation
            returned.  */

int calc_diff(char *s)
{
    int sum, count, i;

    sum = 0;
    count = 0;

    for (i=0; i<msg_size; i++)
        if (msg[i] != s[i])
        {
            count++;
            sum += abs(msg[i] - s[i]);
        }

    if (count == msg_size)
        return sum;
    else
        return count;
}



/*  MIN_DIFF returns the index of the copy that differs least from the target
   string.  If two or more strings tie, then the first one found is selected. */

int min_diff(int *diffs)
{
    int i, index;

    index = 0;
    for (i=1; i<num_copies; i++)
        if (diffs[index] > diffs[i])
            index = i;
    return index;
}



/*  Creates an initial random string.  There is no guarantee that ALL letters
   are different from the corresponding letter in the target.  
   
   Is called repeated for single-step selection.  */

void initialize_string(char *parent, int *diff)
{
    int i;

    memset(parent, '\0', MAX_MSG_SIZE+1);
    for (i=0; i<msg_size; i++)
        parent[i] = letter_pool[Random(pool_size)];
    parent[i] = '\0';
    
    *diff = calc_diff(parent);
}



// (* MATH SUPPORT *)




/*  LOG returns the common log (base 10) of the value, V.  It works for the
   defined REAL_TYPE (Real, Double, or Extended).  The extra testing and
   processing is to get around the range limitations of the LN function. */

real_type Log(real_type v)
{
    int i, count;
    real_type e;
    real_type ln_e;
    real_type x;
  
    if (v == 0.0)
        x = -MAX_VAL;
    else
    {
        if (v > MAX_VAL)
            e = BIG_E;
        else
            if (v < MIN_VAL)
                e = LITTLE_E;
            else
                e = 1.0;
        ln_e = log(e);
        count = 0;
        while ( (v < MIN_VAL) || (v > MAX_VAL) )
        {
            count++;
            v = v / e;
        }
        x = log(v);
        for (i=0; i<count; i++)
            x = x + ln_e;
    }
    return x;
}



/*  POWER takes the value, V, and raises it to the EXPON power, the value of
   which it returns.  It works for the defined REAL_TYPE (Real, Double, or
   Extended).  The extra testing and processing is to get around the range
   limitations of the LN and the EXP function. */

real_type power(real_type v, int expon)
{
    int i, count;
    real_type e;
    real_type ln_e;
    real_type x;

    if (v == 0.0)
        x = 0.0;
    else
    {
        x = expon * log(v);
        if (x > big_ln)
            e = BIG_E;
        else
        {
            if (x < little_ln)
                e = LITTLE_E;
            else
                e = 1.0;
        }
        ln_e = log(e);
        count = 0;
        while ( (x < little_ln) || (x > big_ln) )
        {
            count++;
            x -= ln_e;
        }
        x = exp(x);
        for (i=0; i<count; i++)
            x *= e;
    }
    return x;
}



int Random(int n)
{
    return rand() % n;
}

double random(void)
{
    return (double)rand() / RAND_MAX;
}

